/*
******************************************************************************
*
*      File             : oacs-900.c
*      Version          : 9.0.0, 11. December 2009, Paolo Usai
*      Version          : 8.1.0, 27. March 2009, Karl Hellwig
*                         functional identical to 6.3.0
*      Version          : 7.0.0, 14. June 2007, Paolo Usai,
*      Version          : 6.4.0, 04. June 2007, Karl Hellwig, 
*      Version          : 6.3.0, 02. October 2006, Karl Hellwig
*      Version          : 6.2.0, 30. August 2005, Karl Hellwig
*      Version          : 2.2.0, 27. November 2000, Jochen Junkawitsch
*      Description      : source file for OACS selection rules
*
*
******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "oacs-900.h"


/* ---------------------------------------------------------------------------------------
                                 private functions
   --------------------------------------------------------------------------------------- */

void      selectMode     (CodecMode codecMode, CodecSet* cscs, CodecSet* oacs);
void      notInclude     (CodecMode codecMode, CodecSet* cscs);
int       ready          (CodecSet cscs, CodecSet oacs, int cmacs);
CodecMode findHighest    (CodecSet cs);
CodecMode findLowest     (CodecSet cs);
void      selectLowest   (CodecSet* cscs, CodecSet* oacs);
void      selectHighest  (CodecSet* cscs, CodecSet* oacs);
CodecSet  calc_OACS_FR   (CodecSet cscs, int cmacs);
CodecSet  calc_OACS_HR   (CodecSet cscs, int cmacs);
void      checkDefaultFR (CodecSet* cscs, CodecSet* oacs, int cmacs);
void      checkDefaultHR (CodecSet* cscs, CodecSet* oacs, int cmacs);

/* ---------------------------------------------------------------------------------------

   --------------------------------------------------------------------------------------- */


CodecSet calc_OACS(CodecSet cscs, int cmacs, CodecType codecType)
/*	---------------------------------------------------------------------------------------
	The Codec Type at this point shall be one of FR_AMR or HR_AMR.
	"FR_AMR" comprises here {FR_AMR, UMTS_AMR, UMTS_AMR_2, OHR_AMR}
	"HR_AMR" comprises here {HR_AMR}

	That means: OHR_AMR is not handled separately, but like FR_AMR.
	This is a shortcomming of the TFO Decision Rules:
	The OHR AMR may include the highest AMR Modes as FR_AMR for optimal quality in high C/I.
	But it must be as robust as the HR_AMR for low C/I.
	Anyway: the C-Reference implements what the main TS states.
	--------------------------------------------------------------------------------------- */
{
  switch (codecType)
  {
    case FR_AMR: return calc_OACS_FR(cscs, cmacs);
    case HR_AMR: return calc_OACS_HR(cscs, cmacs);
    default: fprintf(stderr, "ERROR: invalid codec type.\n"); exit(0);
  }
  return (CodecType) 0;
}


CodecSet calc_OACS_FR(CodecSet cscs, int cmacs)
{
  CodecSet oacs = 0;
    
  while (1)
  {
	  
/*	---------------------------------------------------------------------------------------
	Chapter 12.3.1, No HR_Channel is involved, 
	if CMACS == 1
	select mode according to preference list {6.7, 7.4, 5.9, 5.15, 4.75, 7.95, 10.2, 12.2}
	out of the cSCS.
	Note: this leads often to bad results, since the ACSs are not considered.
	The Offerer may heal this to some exted by not offering all modes in their SCS,
	but with the risk of not establishing TFO at all.
	It is strange, that the Acceptability Rule is not applied here.
	--------------------------------------------------------------------------------------- */
	  if (cmacs == 1)
	  {
		  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_102, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_122, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  fprintf(stderr, "ERROR: This point 1 must never be reached.\n"); // exit(0);
	  }
	  
	  
/*	---------------------------------------------------------------------------------------
	Chapter 12.3.1, No HR_Channel is involved, 
	if CMACS == 2
	1.	If mode 10.2 is supported in cSCS, 
		then do not include mode 12.2. => cSCS' = cSCS-12.2
	2.	Select highest mode of cSCS': now we have found one Mode in oACS.

	3.	If mode 12,2 or mode 10,2 has been selected, select second mode according 
		to preference list {6,7, 7,4, 5,9, 5,15, 4,75, 7,95, 10,2, 12,2}.
	4.	Select lowest mode of cSCS. How can we ever get to here?
	--------------------------------------------------------------------------------------- */
	  if (cmacs == 2)
	  {
		  /* If mode 10.2 is supported, do not include mode 12.2. */
		  if (cscs & mode_102) notInclude(mode_122, &cscs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* Select highest mode. */
		  selectHighest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* If mode 12.2 or mode 10.2 is selected, select mode according to
			 preference list {6.7, 7.4, 5.9, 5.15, 4.75, 7.95, 10.2, 12.2}. */
		  if (oacs & (mode_122 | mode_102))
		  {
			  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_102, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
			  selectMode(mode_122, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  }
		  
		  /* Select lowest mode. */
		  selectLowest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  fprintf(stderr, "ERROR: This point 2 must never be reached.\n"); // exit(0);
	  }
	  
/*	---------------------------------------------------------------------------------------
	Chapter 12.3.1, No HR_Channel is involved, 
	if CMACS > 2
	1.	If mode 10,2 is supported, do not include mode 12,2. => cSCS'  = cSCS-12.2
	2.	If mode 4,75 is supported, do not include mode 5,15. => cSCS'' = cSCS'-5.15
	3.	If mode 5,15 is supported, do not include mode 5,9.
	4.	If mode 5,9 is supported and mode 4,75 is not supported, do not include mode 6,7. 
		Note: How can that happen?
	5.	If mode (12,2 or 10,2) and 7,4 is supported, do not include mode 7,95.
	6.	If mode 7,95 is supported, do not include 7,4.

		Note: now we have a strongly modified cSCS''''''''' 
		This may lead to less modes than cMACS, it may even lead to an empty cSCS.
		That is bullshit.....

	7.	Select lowest mode.
	8.	Select highest mode.
	9.	Select mode 6,7.
	10.	Select mode 5,9.

	--------------------------------------------------------------------------------------- */	  
	  
	  /* if 10.2 is supported, do not include 12.2 */
	  if (cscs & mode_102) notInclude(mode_122, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if 4.75 is supported, do not include 5.15 */
	  if (cscs & mode_475) notInclude(mode_515, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.15 is supported, do not include mode 5.9. */
	  if (cscs & mode_515) notInclude(mode_590, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.9 is supported and mode 4.75 is not supported, do not include mode 6.7. */
	  if ((cscs & mode_590) && !(cscs & mode_475)) notInclude(mode_670, &cscs); 
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if (12.2 or 10.2) and 7.4 is supported, do not include 7.95 */
	  if ((cscs & (mode_122 | mode_102)) && (cscs & mode_740)) notInclude(mode_795, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 7.95 is supported, do not include 7.4. */
	  if (cscs & mode_795) notInclude(mode_740, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select lowest mode */
	  selectLowest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select highest mode */
	  selectHighest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select 6.7 */
	  selectMode(mode_670, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select 5.9 */
	  selectMode(mode_590, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  	  
	  /* ----- */
	  
	  fprintf(stderr, "ERROR: This point 3 must never be reached.\n");

  } /* end of while (1) */
  
  if (nModes(oacs) < cmacs)
	  oacs |= cscs;				// fill all remaining Modes of the cSCS into the oACS.

  return oacs;
}


CodecSet calc_OACS_HR(CodecSet cscs, int cmacs)
{
  CodecSet oacs = 0;
    
  while (1)
  {
/*	---------------------------------------------------------------------------------------
	Chapter 12.3.2, A HR_Channel is involved, 
	if CMACS == 1
	Select mode according to preference list {5,9, 5,15, 4,75, 6,7, 7,4, 7,95}
	out of the cSCS.
	Note: this leads often to bad results, since the ACSs are not considered.
	The Offerers may heal this to some exted by not offering all modes in their SCS,
	but with the risk of not establishing TFO at all.
	It is strange, that the Acceptability Rule is not applied here.
	--------------------------------------------------------------------------------------- */	  
	  if (cmacs == 1)
	  {
		  selectMode(mode_590, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_515, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_475, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_670, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_740, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  selectMode(mode_795, &cscs, &oacs); if (ready(cscs, oacs, cmacs)) break;
		  fprintf(stderr, "ERROR: This point 4 must never be reached.\n"); // exit(0);
	  }

/*	---------------------------------------------------------------------------------------
	Chapter 12.3.2, A HR_Channel is involved, 
	if CMACS == 2
	1.	Select highest mode of cSCS
	2.	Select lowest mode of cSCS
	Plain and simple.
	But again the ACSs are not at all considered. Strange.
	--------------------------------------------------------------------------------------- */
	  if (cmacs == 2)
	  {
		  /* Select highest mode. */
		  selectHighest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
		  
		  /* Select lowest mode. */
		  selectLowest(&cscs, &oacs);
		  if (ready(cscs, oacs, cmacs)) break;
 		  fprintf(stderr, "ERROR: This point 5 must never be reached.\n"); // exit(0);
	  }
/*	---------------------------------------------------------------------------------------
	Chapter 12.3.2, A HR_Channel is involved, 
	if CMACS > 2
	1.	If mode 10,2 is supported, do not include mode 12,2. 
		Note: that is obvious for HR
	2.	If mode 4,75 is supported, do not include mode 5,15.
	3.	If mode 5,15 is supported, do not include mode 5,9.
	4.	If mode 5,9 is supported and mode 4,75 is not supported, do not include mode 6,7.
	5.	If mode (12,2 or 10,2) and 7,4 is supported, do not include mode 7,95.
		Note: this can never happen.
	6.	If mode 7,95 is supported, do not include 7,4.

	now we have a modified cSCS''''''''' this may lead to less modes than cMACS.

	7.	Select lowest mode.
	8.	Select highest mode.
	9.	Select mode 6,7.
	10.	Select mode 5,9.

	--------------------------------------------------------------------------------------- */	  

	  /* if 10.2 is supported, do not include 12.2 */
	  if (cscs & mode_102) notInclude(mode_122, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if 4.75 is supported, do not include 5.15 */
	  if (cscs & mode_475) notInclude(mode_515, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.15 is supported, do not include mode 5.9. */
	  if (cscs & mode_515) notInclude(mode_590, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 5.9 is supported and mode 4.75 is not supported, do not include mode 6.7. */
	  if ((cscs & mode_590) && !(cscs & mode_475)) notInclude(mode_670, &cscs); 
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* if (12.2 or 10.2) and 7.4 is supported, do not include 7.95 */
	  if ((cscs & (mode_122 | mode_102)) && (cscs & mode_740)) notInclude(mode_795, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* If mode 7.95 is supported, do not include 7.4. */
	  if (cscs & mode_795) notInclude(mode_740, &cscs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select lowest mode */
	  selectLowest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select highest mode */
	  selectHighest(&cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* ----- */
	  
	  /* select 6.7 */
	  selectMode(mode_670, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  /* select 5.9 */
	  selectMode(mode_590, &cscs, &oacs);
	  if (ready(cscs, oacs, cmacs)) break;
	  
	  
	  /* ----- */
	  
	  fprintf(stderr, "ERROR: This point 6 must never be reached.\n");

  } /* end of while (1) */
  
  if (nModes(oacs) < cmacs)
	  oacs |= cscs;				// fill all remaining Modes of the cSCS into the oACS.

  return oacs;
}


int nModes(CodecSet codecSet)
/*	-------------------------------------------------------
	Counts the Modes in a Codecset
	------------------------------------------------------- */
{
  int k, n=0;
  for (k=0; k<8; k++)
  {
	  if (codecSet & 1) { n++; }
	  codecSet >>= 1;
  }
  return n;
}

void selectMode(CodecMode codecMode, CodecSet* cscs, CodecSet* oacs)
/*	--------------------------------------------------------------------
	Checks, if the desired Mode is still in the cSCS.
	If yes, it deletes the Mode in the cSCS and sets it in the oACS
	-------------------------------------------------------------------- */
{
	if (*cscs & codecMode)					// is this Codec Mode is still in cSCS
	{
		*cscs &= codecMode ^ 0xFF;			// eliminate it in cSCS
		*oacs |= codecMode;					// set it in oACS
	}
}

void notInclude(CodecMode codecMode, CodecSet* cscs)
/*	--------------------------------------------------------------------
	Deletes the Mode in the cSCS
	-------------------------------------------------------------------- */
{
  *cscs &= codecMode ^ 0xFF;
}

int ready(CodecSet cscs, CodecSet oacs, int cmacs)
/*	--------------------------------------------------------------------
	Checks, if a sufficient number of modes is already in the oACS.
	Or: if the remaining modes in the cSCS could just fill the oACS.
	-------------------------------------------------------------------- */
{
    if (nModes(oacs) == cmacs) return 1;
    if (nModes(oacs) + nModes(cscs) <= cmacs) return 1;
    return 0;
}


CodecMode findHighest(CodecSet cs)
{
  if (cs & mode_122) return mode_122;
  if (cs & mode_102) return mode_102;
  if (cs & mode_795) return mode_795;
  if (cs & mode_740) return mode_740;
  if (cs & mode_670) return mode_670;
  if (cs & mode_590) return mode_590;
  if (cs & mode_515) return mode_515;
  if (cs & mode_475) return mode_475;
  // fprintf(stderr, "ERROR: This point 9 must never be reached.\n"); // exit(0);
  // This point CAN be reached in case the the cACS is empty

  return (CodecMode) 0;
}

CodecMode findLowest(CodecSet cs)
{
  if (cs & mode_475) return mode_475;
  if (cs & mode_515) return mode_515;
  if (cs & mode_590) return mode_590;
  if (cs & mode_670) return mode_670;
  if (cs & mode_740) return mode_740;
  if (cs & mode_795) return mode_795;
  if (cs & mode_102) return mode_102;
  if (cs & mode_122) return mode_122;
  // fprintf(stderr, "ERROR: This point 10 must never be reached.\n"); // exit(0);
  // This point CAN be reached in case the the cACS is empty

  return (CodecMode) 0;
}


void selectLowest(CodecSet* cscs, CodecSet* oacs)
{
  selectMode(findLowest(*cscs), cscs, oacs);
}

void selectHighest(CodecSet* cscs, CodecSet* oacs)
{
  selectMode(findHighest(*cscs), cscs, oacs);
}

void checkDefaultFR(CodecSet* cscs, CodecSet* oacs, int cmacs)
/*	-------------------------------------------------------------------------------------------
	The default FR Configuration is defined as          12.2       - 6.7 - 5.9 - 4.75 == Set 12
	Later another important FR Configuration was added: 12.2 - 7.4 -     - 5.9 - 4.75 == Set 1
	------------------------------------------------------------------------------------------- */
{
  if ((cmacs >= 4) &&
      (*cscs & mode_475) &&
      (*cscs & mode_590) &&
      (*cscs & mode_670) &&
      (*cscs & mode_102))
  {
    selectMode(mode_475, cscs, oacs);
    selectMode(mode_590, cscs, oacs);
    selectMode(mode_670, cscs, oacs);
    selectMode(mode_102, cscs, oacs);
  }
}


void checkDefaultHR(CodecSet* cscs, CodecSet* oacs, int cmacs)
/*	-------------------------------------------------------------------------------------------
	The default HR Configuration is defined as                 7.4 - 6.7 - 5.9 - 4.75 == Set 10
	Later another important HR Configuration was added:        7.4 -     - 5.9 - 4.75 == Set 1
	The corresponding important FR Configuration is:    12.2 - 7.4 -     - 5.9 - 4.75 == Set 1
	------------------------------------------------------------------------------------------- */
{
  if ((cmacs >= 4) &&
      (*cscs & mode_475) &&
      (*cscs & mode_590) &&
      (*cscs & mode_670) &&
      (*cscs & mode_740))
  {
    selectMode(mode_475, cscs, oacs);
    selectMode(mode_590, cscs, oacs);
    selectMode(mode_670, cscs, oacs);
    selectMode(mode_740, cscs, oacs);
  }
}


int highModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs)
/*	-----------------------------------------------------------------------------------
	The highModeRule says:
	"Don't give up tandem with high quality modes:
	 The highest mode in the OACS is not lower than one mode 
	 below the minimum of the highest modes of both ACS."
	----------------------------------------------------------------------------------- */
{
  CodecMode highOACS = findHighest(oacs);
  CodecMode highLACS = findHighest(lacs);
  CodecMode highDACS = findHighest(dacs);
  CodecMode minACSMode = (highLACS < highDACS ? highLACS : highDACS);
  
  return (highOACS >= (minACSMode >> 1));
}

int lowModeRule(CodecSet lacs, CodecSet dacs, CodecSet oacs, CodecType codecType)
/*	-----------------------------------------------------------------------------------
	The lowModeRule says: 
	"Tandem AMR with robust low modes performs better, ....
     Either the lowest mode of the oACS is not higher than a specific maximum mode 
	 or both ACS don't contain lower modes than the lowest mode in the OACS. 
	 The specific maximum mode is 5,9 for TFO connections involving a half rate channel 
	 and 7,4 otherwise."
	----------------------------------------------------------------------------------- */
{
  CodecMode max = (codecType == HR_AMR ? mode_590 : mode_740);
  CodecMode lowOACS = findLowest(oacs);
  CodecMode lowLACS = findLowest(lacs);
  CodecMode lowDACS = findLowest(dacs);
  CodecMode lowestACSMode = (lowLACS < lowDACS ? lowLACS : lowDACS);
  
  return (lowOACS <= max || lowOACS <= lowestACSMode);
}

